/*
 * PKCS #11 FIPS Power-Up Self Test - common stuff.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef FIPS_INC
#define FIPS_INC

/* common functions used for FIPS selftests. Due to the modular design of NSS
 * putting these into libfreebl would mean either amending the API represented
 * by FREEBLVectorStr - which might cause problems with newer applications, or
 * extending the API with another similar function set. Thus, to make things
 * less complicated in the binaries, we mess up the source a bit. */

#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>

#include <prtypes.h>
#include <prerror.h>

#include <prlink.h>

#include "blapi.h"

#define NSS_FORCE_FIPS_ENV  "NSS_FIPS"
#define FIPS_PROC_PATH      "/proc/sys/crypto/fips_enabled"
 
#define CHECKSUM_SUFFIX     ".chk"

typedef enum fips_check_status {
    CHECK_UNCHECKED = -1,
    CHECK_OK = 0,
    CHECK_FAIL,
    CHECK_FAIL_CRYPTO,
    CHECK_MISSING
} fips_check_status;

/* initial value of FIPS state is -1 */
static int fips_state = -1;

static int fips_wanted = -1;

/* debug messages are sent to stderr */
static void
debug(const char *fmt,...)
{
#if 0
    va_list args;

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);
    fputc('\n', stderr);
#endif
    return;
}

/* Fatal messages ending with abort(); this function never returns */
static void __attribute__ ((__noreturn__))
fatal(const char *fmt,...)
{
    va_list args;

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);
    fputc('\n', stderr);
    abort();
}

/* check whether FIPS moode is mandated by the kernel */
static int
fips_isWantedProc(void)
{
    int my_fips_wanted = 0;
    int fips_fd;
    char fips_sys = 0;

    struct stat dummy;
    if (-1 == stat(FIPS_PROC_PATH, &dummy)) {
	switch (errno) {
	    case ENOENT:
	    case EACCES:  /* Mozilla sandboxing returns EACCES instead of ENOENT */
	    case ENOTDIR:
		break;
	    default:
		fatal("Check for system-wide FIPS mode is required and %s cannot"
		    " be accessed for reason other than non-existence - aborting"
		    , FIPS_PROC_PATH);
		break;
	}
    } else {
	if (-1 == (fips_fd = open(FIPS_PROC_PATH, O_RDONLY))) {
	    fatal("Check for system-wide FIPS mode is required and %s cannot"
		" be opened for reading - aborting"
		, FIPS_PROC_PATH);
	}
	if (1 > read(fips_fd, &fips_sys, 1)) {
	    fatal("Check for system-wide FIPS mode is required and %s doesn't"
		" return at least one character - aborting"
		, FIPS_PROC_PATH);
	}
	close(fips_fd);
	switch (fips_sys) {
	    case '0':
	    case '1':
		my_fips_wanted = fips_sys - '0';
		break;
	    default:
		fatal("Bogus character %c found in %s - aborting"
		    , fips_sys, FIPS_PROC_PATH);
	}
    }
    return my_fips_wanted;
}

/* "legacy" from lib/sysinit/nsssysinit.c */
static PRBool
getFIPSEnv(void)
{
    char *fipsEnv = getenv("NSS_FIPS");
    if (!fipsEnv) {
	return PR_FALSE;
    }
    if ((strcasecmp(fipsEnv,"fips") == 0) ||
	(strcasecmp(fipsEnv,"true") == 0) ||
	(strcasecmp(fipsEnv,"on") == 0) ||
	(strcasecmp(fipsEnv,"1") == 0)) {
	 return PR_TRUE;
    }
    return PR_FALSE;
}

static int
fips_isWantedEnv(void)
{
    return getFIPSEnv() ? 1 : 0;
}

static int
fips_isWanted(void)
{
    int fips_requests = 0;
#ifdef LINUX
    fips_requests += fips_isWantedProc();
#endif
    if (fips_requests == 0)
        fips_requests += fips_isWantedEnv();
    return fips_requests;
}

/* check integrity signatures (if present) */
static fips_check_status
fips_checkSignature(char *libName, PRFuncPtr addr)
{
    PRBool rv;
    fips_check_status rv_check = CHECK_UNCHECKED;
    int l = PATH_MAX;
    int err = 0;
    int err_NOENT = 0;
    char full_lib_name[PATH_MAX+1];
    full_lib_name[0] = '\0';

    if (NULL == libName) {
	err_NOENT = PR_FILE_NOT_FOUND_ERROR;
	rv = BLAPI_VerifySelf(SHLIB_PREFIX"freebl"SHLIB_VERSION"."SHLIB_SUFFIX, &err);
    } else {
	err_NOENT = PR_FILE_NOT_FOUND_ERROR;
	strncat(full_lib_name, SHLIB_PREFIX, l);
	l -= strlen(SHLIB_PREFIX);
	strncat(full_lib_name, libName, l);
	l -= strlen(libName);
	strncat(full_lib_name, SHLIB_VERSION"."SHLIB_SUFFIX, l);
	l -= strlen(SHLIB_VERSION"."SHLIB_SUFFIX);
#if 1
	rv = BLAPI_SHVerify(full_lib_name, addr, &err);
#else
	rv = 1;
#endif
    }

    if (rv) {
	rv_check = CHECK_OK;
    } else {
	if (err_NOENT == err) {
	    rv_check = CHECK_MISSING;
	} else {
	    rv_check = CHECK_FAIL;
	}
    }

    return rv_check;
}

/* decide what to do depending on the results of tests and system/required FIPS
 * mode */
static int
fips_resolve(fips_check_status check, char *libName)
{
    int state;

    if (fips_wanted) {
	switch (check) {
	    case CHECK_OK:
		debug("fips - %s: mandatory checksum ok"
		    , (libName) ? libName : "freebl");
		break;
	    case CHECK_FAIL:
		fatal("fips - %s: mandatory checksum failed - aborting"
		    , (libName) ? libName : "freebl");
		break;
	    case CHECK_FAIL_CRYPTO:
		fatal("fips - %s: mandatory crypto test failed - aborting"
		    , (libName) ? libName : "freebl");
		break;
	    case CHECK_MISSING:
		fatal("fips - %s: mandatory checksum data missing - aborting"
		    , (libName) ? libName : "freebl");
		break;
	    default:
		fatal("Fatal error: internal error at %s:%u"
		    , __FILE__, __LINE__);
		break;
	}
	state = 1;
    } else {
	switch (check) {
	    case CHECK_OK:
		debug("fips - %s: checksum ok"
		    , (libName) ? libName : "freebl");
		break;
	    case CHECK_FAIL:
#if 0
		fatal("fips - %s: checksum failed - aborting"
		    , (libName) ? libName : "freebl");
#else
		debug("fips - %s: checksum failed - not in FIPS mode; continuing"
		    , (libName) ? libName : "freebl");
#endif
		break;
	    case CHECK_FAIL_CRYPTO:
		fatal("fips - %s: crypto test failed - aborting"
		    , (libName) ? libName : "freebl");
		break;
	    case CHECK_MISSING:
		debug("fips - %s: mandatory checksum data missing, but not required in non FIPS mode; continuing non-FIPS"
		    , (libName) ? libName : "freebl");
		break;
	    default:
		fatal("Fatal error: internal error at %s:%u"
		    , __FILE__, __LINE__);
		break;
	}
	state = 0;
    }
    return state;
}

/* generic selftest
 * libName and addr are the name of shared object to check and a function
 * contained therein; (NULL, NULL) performs selfcheck of freebl.
 * crypto_check is callback that performs cryptographic algorithms checks; NULL
 * for libraries that do not implement any cryptographic algorithms per se
 */
static int
fips_initTest(char *libName, PRFuncPtr addr, fips_check_status cryptoCheck(void))
{
    fips_check_status check = CHECK_OK;

    fips_wanted = fips_isWanted();

    if (cryptoCheck) {
	check = cryptoCheck();
        debug("fips - %s: crypto check %s"
	    , (libName) ? libName : "freebl"
	    , (CHECK_OK == check) ? "ok" : "failed");
    }

    if (CHECK_OK == check) {
	check = fips_checkSignature(libName, addr);
    }

    return fips_resolve(check, libName);
}

#endif
